#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <cbm.h>
#include <peekpoke.h>
#include <string.h>
#include <6502.h>

#define p_clr  0x93 /* Schoon scherm */
#define p_hlb  0xb0 /* Hoek linksboven */
#define p_hrb  0xae /* Hoek rechtsboven */
#define p_hlo  0xad /* Hoek linksonder */
#define p_hro  0xbd /* Hoek rechtsonder */
#define p_hor  0x60 /* Steep horizontaal */
#define p_ver  0x7d /* Streep verticaal */

#ifdef __C64__
#define TAPE_BUFFER ((char *) 0x033c)
#define BANK8K
#define bankmask     0x1fff
#define basic_start  0x0800
#define tool_size_l  PEEKW(0x087e)
#define tool_size_h  PEEKW(0x08fe)
#endif
#ifdef __C128__
#define TAPE_BUFFER ((char *) 0x0b00)
#define BANK16K
#define bankmask     0x3fff
#define basic_start  0x1c00
#define tool_size_l  PEEKW(0x1c7e)
#define tool_size_h  PEEKW(0x1cfe)
#endif

#define SET_ST(x) { POKE (0x90,x); }
#define GET_ST() (PEEK(0x90))

#define TI_HI 160
#define TI_MID 161
#define TI_LO 162

#define GMOD2_BANK(x) { POKE (0xde00,x); }
#define GMOD2_DATA()  (PEEK(0xde00))

char		device,olddevice;
char		carttype,eeprom_bit; /* 0 - none   1 - Gmod2   2 - Magic Desk   255 - unsuppported */
char		filename[32];
char		*filename_extra;
char		secprot[8];
char		flashidx,flashsecsz,flashseccnt;
char		fast1541;
unsigned	seca,secz;

/* Flash memories with sector protect should come before
   ones without. */
#define flash_idx_AM29F040		0
#define flash_idx_A29040B		1
#define flash_idx_EN29F040		2
#define flash_idx_HY29F040		3
#define flash_idx_M29F040		4
#define flash_idx_MBM29F040		5
#define flash_idx_MX29F040		6
#define flash_idx_SST39SF040	7
#define flash_idx_SST39VF040	8
#define flash_idx_W39L040		9

/* Highest idx that has sector protect */
#define flash_idx_last_with_protect flash_idx_MBM29F040

#define flash_chips_count 10

#ifdef BUS_REVERSED
unsigned	flashrom_ids[] = {
	0x8025, /* AM29F040 (TMS29F040, AS29F040)  */
	0xec61, /* A29040B (AS29CF040) */
	0x3820, /* EN29F040 */
	0xb502, /* HY29F040 */
	0x0447, /* M29F040 */
	0x2025, /* MBM29F040 */
	0x4325, /* MX29F040C */
	0xfded, /* SST39SF040 */
	0xfdeb, /* SST39VF040 */
	0x5b6b  /* W39L040 */
};
#else
unsigned	flashrom_ids[] = {
	0x01a4, /* AM29F040 (TMS29F040, AS29F040) */
	0x3786, /* A29040B (AS29CF040) */
	0x1c04, /* EN29F040 */
	0xad40, /* HY29F040, BM29F040 */
	0x20e2, /* M29F040 */
	0x04a4, /* MBM29F040 */
	0xc2a4, /* MX29F040C */
	0xbfb7, /* SST39SF040 */
	0xbfd7, /* SST39VF040 */
	0xdad6  /* W39L040 */
};
#endif

/* 20e3 M29W040
   52a4 AS29F040
   014f AM29LV040
   3792 A29L040
   1f13 AT49F040
   1c4f EN29LV040
   9d4e PM39F040
   bf13 SST29SF040
*/

const char *flashrom_names[] = {
	"AM29F040",
	"A29040B",
	"EN29F040",
	"HY29F040",
	"M29F040",
	"MBM29F040",
	"MX29F040",
	"SST39SF040",
	"SST39VF040",
	"W39L040"
};

/* 1 << secsz == sector size */
const char secsz[] = {
	16, /* AM29F040 */
	16, /* A29C040B */
	16, /* EN29F040 */
	16, /* HY29F040 */
	16, /* M29F040 */
	16, /* MBM29F040 */
	16, /* MX29F040C */
	12, /* SST39SF040 */
	12, /* SST39VF040 */
	16  /* W39L040 */
};

const unsigned seccnt[] = {
	8, /* AM29F040 */
	8, /* A29040B */
	8, /* EN29F040A */
	8, /* HY29F040 */
	8, /* M29F040 */
	8, /* MBM29F040 */
	8, /* MX29F040C */
	128, /* SST39SF040 */
	128, /* SST39VF040 */
	8  /* W39L040 */
};

extern void strout(register const char *s);
/*void strout(register const char *s) {
  while (*s!=0) {
    cbm_k_bsout(*s);
    ++s;
  }
}*/

void numout(unsigned v) {
  utoa(v,TAPE_BUFFER,10);
  strout(TAPE_BUFFER);
}

extern void serialout(register const char *s);
/*
void serialout(register const char *s) {
  while (*s!=0) {
    cbm_k_ciout(*s);
    ++s;
  }
}*/

void serial_skip(char n);
/*void serial_skip(char n) {
  do {
    cbm_k_acptr();
    n--;
  } while (n!=0);
}*/

void fast1541_disable_screen() {
  if (fast1541) {
    POKE(0xd011,PEEK(0xd011) & 0xef);
    /* The VIC-II decides at raster line 48 wether or not to generate bad
       lines and no bad lines occur at lines before*/
    SEI();
    do {} while (PEEK(0xd012)|(PEEK(0xd011) & 0x80)<=48);
    CLI();
  }
}

void fast1541_enable_screen() {
  if (fast1541)
    POKE(0xd011,PEEK(0xd011) | 0x10);
}

char send_drive_command(const char *s) {
  static char c;
  SET_ST(0);
  cbm_k_listen(device);
  cbm_k_second(0x6f);
  if (c=GET_ST()) {
    cbm_k_unlsn();
    return c;
  }
  serialout(s);
  cbm_k_unlsn();
  return GET_ST();
}

void print_drive_status() {
  /* Device MUST be present, checking st doesn't help after TALK */
  register char *s;

  SET_ST(0);
  fast1541_disable_screen();
  cbm_k_talk(device);
  cbm_k_tksa(0x6f);
  s=TAPE_BUFFER;
  do {
    *s=cbm_k_acptr();
  } while (*s++!=13);
  *s=0;
  cbm_k_untlk();
  fast1541_enable_screen();
  strout(TAPE_BUFFER);
}

void handle_st(const char st) {
  if (st & 128)
    strout("Device not present");
  if (st & 64)
    strout("End of file");
  if (st & 16)
    strout("Verify");
  if (st & 2)
    strout("Read");
  if (st & 1)
    strout("Write ");
  if (st)
    strout(" error.\n");
}

void file_open(const char *s, const char chan) {
  cbm_k_listen(device);
  cbm_k_second(0xf0 | chan);
  serialout(s);
  cbm_k_unlsn();
}

void file_close(const char chan) {
 /* Tell device we want to close the directory. */
  cbm_k_listen(device);
  cbm_k_second(0xe0 | chan);
  cbm_k_unlsn();
}

void directory(const char *s) {
  static unsigned size;
  static char c;

  file_open(s,0);  /* Chan 0 = load PRG */
 
  SET_ST(0);
  /* Make device talk the directory. */
  fast1541_disable_screen();
  cbm_k_talk(device);
  cbm_k_tksa(0x60);  /* Chan 0 = load PRG */
  /* Skip load address */
  serial_skip(2);
  do {
    /* Skip BASIC line pointer */
    serial_skip(2);
    if (GET_ST())
      break;
    /* Compute the file size in blocks and print it. */
    size = cbm_k_acptr() | (cbm_k_acptr() << 8);
    if (GET_ST())
      break;
    numout(size);
    cbm_k_bsout(' ');

    /* Get the rest of the directory entry and print it */
    do {
      c = cbm_k_acptr();
      if (c==0 || GET_ST())
        break;
      cbm_k_bsout(c);
    } while (1);
    if (GET_ST())
      break;
    cbm_k_bsout(13);
  } while (1);
  cbm_k_untlk();
  fast1541_enable_screen();

  file_close(0);
}

void handle_wildcard() {
  static char *s;

  filename[0]='$';
  file_open(filename,0); /* Chan 0 = load PRG */
 
  SET_ST(0);
  /* Make device talk the directory. */
  fast1541_disable_screen();
  cbm_k_talk(device);
  cbm_k_tksa(0x60);  /* Chan 0 = load PRG */
  /* Skip load address, BASIC line pointer and line number (file size)*/
  serial_skip(6);
  do {} while (cbm_k_acptr()!=0);
  /* Skip BASIC line pointer and line number (file size)*/
  serial_skip(4);
  do {} while (cbm_k_acptr()!='"');
  s=filename+1;
  do {} while ((*s++=cbm_k_acptr())!='"');
  s--;
  filename_extra=s;
  *s=0;
  cbm_k_untlk();
  fast1541_enable_screen();

  file_close(0);
}

void set_filename(register const char *s) {
  register char *t;
  static char wildcard;

  wildcard=0;
  t=filename+1;
  while ((*t=*s)!=13) {
    if (*s=='*')
      wildcard=1;
    ++s;
    ++t;
  }
  filename_extra=t;
  *t=0;

  if (wildcard)
    handle_wildcard();

  strout("Setting filename to \"");
  strout(filename+1);
  strout("\"\n");
}

//#define cursor(b) {if (b) POKE(204,0); else POKE(204,255);}

extern void swap_stubs();
/*
void swap_stubs() {
  register char *s,*t,c;
  s=(char *)0x800;
  t=(char *)0x880;
  do {
    c=*s;
    *s=*t;
    *t=c;
    ++s;
    ++t;
  } while (((unsigned)t & 0xff)!=0);
}*/


void dump_tool_to_file() {
  register unsigned s;
  static unsigned t;

  SET_ST(0);

#ifdef __C64__
  if (PEEK(0x0884)!=0xc3)  /* C from CBM80 */
    swap_stubs();
#endif
#ifdef __C128__
  if (PEEK(0x1c87)!=0x43)  /* C from CBM */
    swap_stubs();
#endif

  file_open(filename+1,1);   /* Chan 1 = save PRG */
  if (GET_ST())
    goto error;

  /* Give listen command */
  cbm_k_listen(device);
  cbm_k_second(0x61);  /* Chan 1 = save PRG */
  if (GET_ST())
    goto error;

  /* Write load address 0x0801 */
  cbm_k_ciout(0x01);
  cbm_k_ciout(basic_start>>8);

  /* Write content */
  s=basic_start+1;
  t=s+tool_size_l-1;
  do  {
    cbm_k_ciout(PEEK(s));
    ++s;
  } while (s!=t);
  if (GET_ST())
    goto error;
  cbm_k_unlsn();

  strout("File written.\n");
  goto end;

error:
  handle_st(GET_ST());
//  strout("Failed to write file.\n");
end:
  file_close(1);
  filename_extra[0]=0;
}

#ifdef __C64__
char test_roml_ram() {
  POKE(0x8000,0);
  if (PEEK(0x8000)!=0)
    return 0;
  POKE(0x8000,0xff);
  if (PEEK(0x8000)!=0xff)
    return 0;
  return 1;
}

char gmod2_detect() {
  /* Returns:
     0 - No cart (ROML is RAM)
     1 - Gmod2
     2 - Magic desk
     3 - Unknown cart (ROML stuck at ROM)
   */
  static char c;
  /* Assume EXROM active*/
  if (test_roml_ram())
    return 0;
  /* Deactivate EXROM for Gmod2*/
  GMOD2_BANK(64);
  c=1;
  if (!test_roml_ram()) {
    /* Deactivate EXROM for Magic Desk*/
    GMOD2_BANK(128);
    if (test_roml_ram())
      c=2;
    else
      c=0xff;
  };
  /* Reactivate EXROM */
  GMOD2_BANK(0);
  return c;
}
#endif

void flash_magic_unlock() {
  /* Flash magic address 1: 0x5555 = bank 2, offset 0x1555 */
  /* Flash magic address 2: 0x2aaa = bank 1, offset 0x0aaa */
#ifdef BANK8K
#ifdef BUS_REVERSED
  GMOD2_BANK(0xc2);
  POKE(0xe000+0x1555,0x55);
  GMOD2_BANK(0xc1);
  POKE(0xe000+0x0aaa,0xaa);
#else
  GMOD2_BANK(0xc2);
  POKE(0xe000+0x1555,0xaa);
  GMOD2_BANK(0xc1);
  POKE(0xe000+0x0aaa,0x55);
#endif
#else
  /* BANK16K */
  GMOD2_BANK(0xc1);
  POKE(0x8000+0x1555,0xaa);
  GMOD2_BANK(0xc0);
  POKE(0x8000+0x2aaa,0x55);
#endif
}

void flash_mode_autoselect() {
  /* put the flash ROM in its "autoselect" mode */
  flash_magic_unlock();
#ifdef BANK8K
#ifdef BUS_REVERSED
  GMOD2_BANK(0xc2);
  POKE(0xe000+0x1555,0x09);
#else
  GMOD2_BANK(0xc2);
  POKE(0xe000+0x1555,0x90);
#endif
#else
  /* BANK16K */
  GMOD2_BANK(0xc1);
  POKE(0x8000+0x1555,0x90);
#endif
}

void flash_mode_reset() {
  GMOD2_BANK(0xc0);
#ifdef BANK8K
#ifdef BUS_REVERSED
  POKE(0xe000,0x0f);
#else
  POKE(0xe000,0xf0);
#endif
#else
  /* BANK16K */
  POKE(0x8000,0xf0);
#endif
}

char flash_write_byte(const char bank, const unsigned offset, const char c) {
  static char count;

  flash_magic_unlock();
#ifdef BANK8K
  GMOD2_BANK(0xc2);
#ifdef BUS_REVERSED
  POKE(0xe000+0x1555,0x05);
#else
  POKE(0xe000+0x1555,0xa0);
#endif
#else
  /* BANK16K */
  GMOD2_BANK(0xc1);
  POKE(0x8000+0x1555,0xa0);
#endif
  GMOD2_BANK(bank | 0xc0);
#ifdef BANK8K
  POKE(0xe000 | offset,c);
#else
  POKE(0x8000 | offset,c);
#endif
  /* C128 can read flash if writing enabled */
#ifdef __C64__
  GMOD2_BANK(bank);
#endif
  count=192;
  while (PEEK(0x8000 | offset)!=c) {
    if (++count == 0)
      return 0;
  }
  return 1;
}

char flash_erase_sector(char sector) {
  static char v;
  static char count;
  static char offset,bank;

  strout("Erasing sector ");
  numout(sector);
  cbm_k_bsout('\n');
  /* Critical timing, delay of interrupt in the middle unacceptable */
  SEI();
  flash_magic_unlock();
#ifdef BANK8K
  GMOD2_BANK(0xc2);
#ifdef BUS_REVERSED
  POKE(0xe000+0x1555,0x01);
#else
  POKE(0xe000+0x1555,0x80);
#endif
#else
  /* BANK16K */
  GMOD2_BANK(0xc1);
  POKE(0x8000+0x1555,0x80);
#endif
  flash_magic_unlock();
  if (flashsecsz==12) {
#ifdef BANK8K
    offset=0;
    if (sector & 1)
      offset=0x10;
    bank=sector>>1;
#else
    offset=(sector & 3) << 4;
    bank=sector>>2;
#endif
  } else {
    /* 16 */
    offset=0;
#ifdef BANK8K
    bank=sector<<3;
#else
    bank=sector<<2;
#endif
  }
  GMOD2_BANK(0xc0|bank);
#ifdef BANK8K
#ifdef BUS_REVERSED
  POKE(0xe000|(offset<<8),0x0c);
#else
  POKE(0xe000|(offset<<8),0x30);
#endif
#else
  POKE(0x8000|(offset<<8),0x30);
#endif
  CLI();
  /* C128 can read flash if writing enabled */
#ifdef __C64__
  GMOD2_BANK(bank);
#endif
  /* Wait at least 8 seconds */
  count=252;
retry:
  /* TI clock middle byte changes every 255/60 = 4.25 sec on 60Hz */
  v = PEEK(TI_MID);

  do {
    if (PEEK(TI_MID)!=v) {
      if (++count==0)  /* 60 Hz = worst case */
        return 0;  /* TImeout */
      goto retry;
    }
  } while (PEEK(0x8000|(offset<<8))!=255);
  return 1;
}

void write_tool_to_flash() {
  register unsigned s = 0;

#ifdef BANK8K
  if (!flash_erase_sector(0))
    goto error;
  if (flashsecsz<=12)
    if (!flash_erase_sector(1))
      goto error;
#else
  register char i,z;
  z=1;
  if (flashsecsz<=12)
    z=4;
  for (i=0;i<z;i++)
    if (!flash_erase_sector(0))
      goto error;
#endif
  /* Make sure cartridge stub comes first */
#ifdef __C64__
  if (PEEK(0x0804)!=0xc3)  /* C from CBM80 */
    swap_stubs();
#endif
#ifdef __C128__
  if (PEEK(0x1c07)!=0x43)  /* C from CBM */
    swap_stubs();
#endif
  strout("Writing tool\n");
  do {
    if (!flash_write_byte(0,s,PEEK(basic_start+s)))
      goto error;
  } while (++s!=tool_size_h);
  return;
error:
  strout("Flash write error\n");
  numout(s);
}


void flash_detect() {
   static unsigned flashid;
   register char i;

  flash_mode_autoselect();
#ifdef __C64__
  GMOD2_BANK(0);
#endif
  flashid=(PEEK(0x8000)<<8) | PEEK(0x8001);
  cbm_k_bsout(13);
  for (i=4;i!=0;i--) {
    flash_mode_reset();
    flash_mode_autoselect();
#ifdef BANK8K
    GMOD2_BANK(i<<3);  /* Swap banks to avoid collision with memory contents */
#else
    GMOD2_BANK(i<<2);  /* Swap banks to avoid collision with memory contents */
#endif
    if (flashid!=((PEEK(0x8000)<<8) | PEEK(0x8001))) {
      cbm_k_bsout(13);
      strout("Write access unreliable error.\n");
      flash_mode_reset();
      return;
    }
  };
/*  strout("Flash ID: ");
  numout(flashid);
  cbm_k_bsout('\n');*/
  flashidx=0xff;
  for (i=0;i<flash_chips_count;i++)
    if (flashrom_ids[i]==flashid) {
      flashidx=i;
      break;
    }
  for (i=0;i<8;i++)
    secprot[i]=0;
  if (flashidx!=0xff) {
    strout(flashrom_names[flashidx]);
    if (flashidx<=flash_idx_last_with_protect) {
      strout(", writeable sectors ");
      for (i=0;i<8;i++) {
        GMOD2_BANK(i<<3);
        secprot[i]=PEEK(0x8002);
        if (PEEK(0x8002))
          cbm_k_bsout('-');
        else
          cbm_k_bsout(186);  /* Checkmark*/
      };
    }
    flashsecsz=secsz[flashidx];
    cbm_k_bsout('\n');
    strout("Sector cnt: ");
    numout(flashseccnt);
    strout(" size: 2^");
    numout(flashsecsz);
    cbm_k_bsout('\n');

    if (flashseccnt!=seccnt[flashidx]) {
      flashseccnt=seccnt[flashidx];
      strout("seca=0 secz=");
      seca=0;
      secz=flashseccnt-1;
      numout(secz);
    }
  } else {
    strout("No supported flash found.");
  }
  cbm_k_bsout('\n');
  flash_mode_reset();
  GMOD2_BANK(0);
}

void filename_set_extra(char c) {
  filename_extra[0]=',';
  filename_extra[1]='s';
  filename_extra[2]=',';
  filename_extra[3]=c;
  filename_extra[4]=0;
}

void flash_write_file() {
  register unsigned cursec,secptr,secend;
  static unsigned base;
  static char bank,b;

  filename_set_extra('r');
  SET_ST(0);
  file_open(filename+1,2);
  if (GET_ST())
    goto file_error;
  fast1541_disable_screen();
  cbm_k_talk(device);
  cbm_k_tksa(0x62);
  if (GET_ST())
    goto file_error;
  for (cursec=seca;cursec<=secz;cursec++) {
    if (!flash_erase_sector(cursec))
      goto flash_error;
    strout("Writing from file to flash sector ");
    numout(cursec);
    cbm_k_bsout('\n');
    secptr=0;
    base=0;
    if (flashsecsz==12) {
#ifdef BANK8K
      if (cursec & 1) {
        base=0x1000;
      }
      bank=cursec>>1;
#else
      base=(cursec & 3) << 12;
      bank=cursec>>2;
#endif
      secend=0x1000;
    } else {
      /* 16 */
      secend=0;
#ifdef BANK8K
      bank=(cursec<<3) | (secptr>>13);
#else
      bank=(cursec<<2) | (secptr>>14);
#endif
    }
    do {
      if ((secptr & bankmask) == 0) {
        if (flashsecsz==12) {
#ifdef BANK8K
          bank=cursec>>1;
#else
          bank=cursec>>2;
#endif
        } else {
#ifdef BANK8K
          bank=(cursec<<3) | (secptr>>13);
#else
          bank=(cursec<<2) | (secptr>>14);
#endif
        }
      }
      b=cbm_k_acptr();
      if (GET_ST() && (secptr!=(secend-1)))
        goto file_error;
      if (!flash_write_byte(bank,base|(secptr & bankmask),b))
        goto flash_error;
      ++secptr;
    } while (secptr!=secend);
  }
  goto end;
file_error:
  handle_st(GET_ST());
  goto end;
flash_error:
  strout("Flash write error\n");
  numout(bank);
  cbm_k_bsout('\n');
  numout(secptr);
  cbm_k_bsout('\n');
end:
  cbm_k_untlk();
  fast1541_enable_screen();
  filename_extra[0]=0;
  file_close(2);
}


void flash_dump_to_file() {
  register unsigned cursec,secptr,secend;
  static char bank;

  if (carttype!=1 && carttype!=2) {
    strout("No cart error.\n");
    return;
  }
  filename_set_extra('w');
  file_open(filename+1,2);
  cbm_k_listen(device);
  cbm_k_second(0x62);
  for (cursec=seca;cursec<=secz;cursec++) {
    strout("Writing flash sector ");
    numout(cursec);
    strout(" to file.\n");
    secptr=0;
    if (flashsecsz==12) {
#ifdef BANK8K
      if (cursec & 1) {
        secptr=0x1000;
      }
      bank=cursec>>1;
#else
      secptr=(cursec & 3) << 12;
      bank=cursec>>2;
#endif
      secend=0x1000;
    } else {
      /* 16 */
      secend=0;
#ifdef BANK8K
      bank=(cursec<<3) | (secptr>>13);
#else
      bank=(cursec<<2) | (secptr>>14);
#endif
    }
    GMOD2_BANK(bank);
    do {
      cbm_k_ciout(PEEK(0x8000 | (secptr & bankmask)));
      ++secptr;
      if ((secptr & bankmask) == 0) {
        if (flashsecsz==16) {
          bank++;
          GMOD2_BANK(bank);
        }
      }
    } while (secptr!=secend);
  }
  cbm_k_unlsn();
  filename_extra[0]=0;
  file_close(2);
}

void eeprom_opcode(const char o) {
  static char v;
  /* Assuming S is not yet active */
  /* Send 1, the start bit */
  GMOD2_BANK(eeprom_bit|16);            /* S=1 D=1  C=0*/
  GMOD2_BANK(eeprom_bit|32|16);         /* S=1 D=1  C=1*/ 
  /* Send bit 7 */
  v=(o & 0x80)>>3;
  GMOD2_BANK(eeprom_bit|v);             /* S=1 D=o7 C=0*/
  GMOD2_BANK(eeprom_bit|32|v);          /* S=1 D=o7 C=1*/
  /* Send bit 6 */
  v=(o & 0x40)>>2;
  GMOD2_BANK(eeprom_bit|v);             /* S=1 D=06 C=0*/
  GMOD2_BANK(eeprom_bit|32|v);          /* S=1 D=06 C=1*/
}

void eeprom_addr(const unsigned addr) {
  static char v;
  static unsigned m;

  m=512;
  do {
    v=eeprom_bit;
    if (addr & m)
      v|=16;
    GMOD2_BANK(v);
    v|=32;
    GMOD2_BANK(v);
    m>>=1;
  } while (m!=0);
}

unsigned eeprom_data_read() {
  static unsigned m;
  static unsigned data;

  m=0x8000;
  data=0;
  /* Clock low so chip outputs dummy bit */
  GMOD2_BANK(eeprom_bit);
  do {
    /* Clock high */
    GMOD2_BANK(eeprom_bit|32);
    /* Clock low */
    GMOD2_BANK(eeprom_bit);
    /* If GMOD2_BANK is followed by GMOD2_DATA, the compiler will
       optimize lda after sta to the same memory location away.
       CC65 does not support "volatile", so we have to confuse the
       optimizer. A "nop" seems to do the trick. */
    asm("nop");
    if ((GMOD2_DATA() & 0x80)!=0)
      data|=m;
    m>>=1;
  } while (m != 0);
  /* Deselect chip */
  GMOD2_BANK(0);
  return data;
}

void eeprom_data_write(const unsigned data) {
  static char v;
  static unsigned m;

  m=0x8000;
  do {
    v=eeprom_bit;
    if (data & m)
      v|=16;
    GMOD2_BANK(v); /* Present data */
    v|=32;
    GMOD2_BANK(v); /* Add clock to it. */
    m>>=1;
  } while (m != 0);
  /* Deselect clock & deselect chip */
  GMOD2_BANK(0);
  /* Reselect chip */
  GMOD2_BANK(eeprom_bit);
  /* Wait until write completed. */
  do {} while ((GMOD2_DATA() & 0x80)==0);
  /* Deselect chip */
  GMOD2_BANK(0);
}

void dump_eeprom_to_file() {
  static unsigned i,d;

  SET_ST(0);

  filename_set_extra('w');
  file_open(filename+1,2);   /* Chan 1 = save PRG */
  if (GET_ST())
    goto error;

  /* Give listen command */
  cbm_k_listen(device);
  cbm_k_second(0x62);  /* Chan 1 = save PRG */
  if (GET_ST())
    goto error;

  /* Write content */
  for (i=0;i<1024;i++) {
    eeprom_opcode(0x80);
    eeprom_addr(i);
    d=eeprom_data_read();
    cbm_k_ciout(d >> 8);
    cbm_k_ciout(d & 0xff);
  }
  if (GET_ST())
    goto error;
  cbm_k_unlsn();

  strout("File written.\n");
  goto end;

error:
  handle_st(GET_ST());
end:
  file_close(2);
  filename_extra[0]=0;
}

void eeprom_write_file() {
  static unsigned i,d;

  SET_ST(0);

  filename_set_extra('r');
  file_open(filename+1,2);
  if (GET_ST())
    goto error;

  /* Give talk command */
  fast1541_disable_screen();
  cbm_k_talk(device);
  cbm_k_tksa(0x62);
  if (GET_ST())
    goto error;

  /* Enable writing on EEPROM */
  eeprom_opcode(0);
  eeprom_addr(768);
  GMOD2_BANK(0);

  /* Write content */
  for (i=0;i<1024;i++) {
    d=(cbm_k_acptr() << 8) | cbm_k_acptr();
    if (GET_ST() && (i!=1023))
      goto error;
    eeprom_opcode(0x40);
    eeprom_addr(i);
    eeprom_data_write(d);
  }
  cbm_k_untlk();

  goto end;

error:
  handle_st(GET_ST());
end:
  fast1541_enable_screen();
  /* Disable writing on EEPROM */
  eeprom_opcode(0);
  eeprom_addr(0);
  GMOD2_BANK(0);

  file_close(2);
  filename_extra[0]=0;
}

void print_header() {
  static const char header[] = {
    p_clr,
    p_hlb,p_hor,p_hor,p_hor,p_hor,p_hor,p_hor,p_hor,p_hor,p_hor,p_hor,p_hor,p_hor,p_hor,p_hrb,  13,
    p_ver,'G'  ,'m',  'o',  'd',  '2',  ' ',  'f',  'l',  'a',  's',  'h',  'e',  'r'  ,p_ver,  13,
    p_hlo,p_hor,p_hor,p_hor,p_hor,p_hor,p_hor,p_hor,p_hor,p_hor,p_hor,p_hor,p_hor,p_hor,p_hro,  13,
    13,0
  };

  strout(header);
  strout("(C) 2021 by Daniel Mantione\n\n"
         "Type \"help\" for help.\n");
}

void print_help() {
  strout("Commands you enter are passed to the de-"
         "vice as DOS commands. To switch between\n"
         "devices, enter their device number.\n\n");
  strout("Special commands:\n"
          "* = wildcard supported\n"
          "help    - This help text.\n"
          "$       - directory*\n"
          "file:fn - Set \"fn\" as file to flash*\n"
          "seca:n  - \"n\" = first sector to flash\n"
          "secz:n  - \"n\" = last sector to flash\n"
          "edump   - Dump eeprom to file\n"
          "fdump   - Dump flash to file\n"
          "tdump   - Dump this tool to file\n");
  if (flashidx!=0xff) {
    strout("ewrit   - Write eeprom to flash\n"
           "fwrit   - Write file to flash\n"
           "twrit   - Write this tool to flash\n");
  } else {
    strout("won     - Enable write commands\n");
  }
#ifdef __C64__
  if(PEEK(0x8a)!=0x60) /* Check CHRGET presence */
    strout("exit    - Boot to BASIC\n");
  else
#endif
#ifdef __C128__
  if(PEEK(0x0ac0)!=0xff) /* Active function ROM number */
    strout("exit    - Boot to BASIC\n");
  else
#endif
    strout("exit    - Exit to BASIC\n");
}

void input_command() {
  register char *s;
  numout(device & 0x7f);
  if (device & 0x80)
    strout("!!");
  strout("> ");
  s=TAPE_BUFFER;
  do {
    *s=cbm_k_basin();
  } while (*s++!=13);
  *s=0;
}

void main_loop() {
//  const char *s = TAPE_BUFFER;
  register char *s,*t;
  do {
    if (device!=olddevice) {
      char c;
      c=send_drive_command("ui");
      handle_st(c);
      device|=c & 0x80;
      olddevice=device;
    }
    if ((device & 0x80) == 0)
      print_drive_status();
    input_command();
    cbm_k_bsout(13);
    s=TAPE_BUFFER;
    while (*s==' ')
      s++;
    /* Directory */
    if (s[0]=='$' && (device & 0x80) == 0) {
      t=s;
      while (*t!=13)
        ++t;
      *t=0;
      directory(s);
    }
    /* Device ID*/
    else if (s[0]>='0' && s[0]<='9' && s[1]==13 && !fast1541)
      device=s[0]-'0';
    else if (((s[0]>='0' && s[0]<='2' && s[1]>='0' && s[1]<='9') || (s[0]=='3' && s[1]=='0')) && s[2]==13 && !fast1541)
      device=10*(s[0]-'0')+(s[1]-'0');
    /* Exit */
    else if (s[0]=='e') { 
      if (s[1]=='x' && s[2]=='i' && s[3]=='t')
        break;
      if (s[1]=='d' && s[2] =='u' && s[3]=='m' && s[4]=='p')
        dump_eeprom_to_file();
      if (s[1]=='w' && s[2] =='r' && s[3]=='i' && s[4]=='t')
        eeprom_write_file();
    } else if (s[0]=='f' && (device & 0x80) == 0) {
      /* fdump */
      if (s[1]=='d' && s[2] =='u' && s[3]=='m' && s[4]=='p')
        flash_dump_to_file();
      /* File */
      if (s[1]=='i' && s[2] =='l' && s[3]=='e' && s[4]==':')
        set_filename(s+5);
      /* fwrit */
      if (s[1]=='w' && s[2] =='r' && s[3]=='i' && s[4]=='t' && flashidx!=0xff)
        flash_write_file();
    }
    /* Help */
    else if (s[0]=='h' && s[1]=='e' && s[2]=='l' && s[3]=='p')
      print_help();
    else if (s[0]=='s' && s[1]=='e' && s[2]=='c' && s[4]==':') {
      if (s[3]=='a') {
        /* seca */
        seca=atoi(s+5);
        strout("Start");
      } else {
        /* secz */
        secz=atoi(s+5);
        strout("End");
      };
      strout(" sector set to ");
      strout(s+5);
    } else if (s[0]=='t' && (device & 0x80) == 0) {
      /* Dump tool to file */
      if (s[1]=='d' && s[2] =='u' && s[3]=='m' && s[4]=='p')
        dump_tool_to_file();
      if (s[1]=='w' && s[2] =='r' && s[3]=='i' && s[4]=='t' && flashidx!=0xff)
        write_tool_to_flash();
    } else if (s[0]=='w' && s[1]=='o' && s[2]=='n') {
      flash_detect();
    } else
      if (device & 0x80)
        handle_st(device & 0x80);
      else
        send_drive_command(s);
      if (s[0]=='u' && s[1]=='i') {
        if (s[2]=='-')
          fast1541=1;
        if (s[2]=='+')
          fast1541=0;
      }
  } while (1);
}

int main (void) {
#ifdef __C64__
  /* Enable ROM */
  POKE(1,7);
#endif
#ifdef __C128__
  /* $0000..$7fff = RAM Bank 0
     $8000..$bfff = Cartridge ROML
     $c000..$cfff = Screen editor
     $d000..$dfff = IO
     $e000..$ffff = KERNAL */
  POKE(0xff00,0x0a);
#endif
  fast1541=0;
  olddevice=0;
  device=8;
  flashidx=255;
  seca=0;
  secz=7;
  flashseccnt=8;
  print_header();
#ifdef __C64__
  strout("Banksw reg behaves as: ");
  carttype=gmod2_detect();
  switch (carttype) {
    case 0:
      strout("No cart inserted\n");
      break;
    case 1:
      strout("Gmod2\n");
      eeprom_bit=64;
      break;
    case 2:
      strout("Magic desk\n");
      eeprom_bit=128;
    case 255:
      strout("Unsupported cart\n");
  }
#endif
#ifdef __C128__
  /* C128 does not allow cartridge to influence memory config */
  eeprom_bit=64;
#endif

  main_loop();
  /* Reset cartridge. */
  GMOD2_BANK(0);
  /* BASIC stub first */
#ifdef __C64__
  if (PEEK(0x0884)!=0xc3)  /* C from CBM80 */
    swap_stubs();
#endif
#ifdef __C128__
  if (PEEK(0x1c87)!=0x43)  /* C from CBM */
    swap_stubs();
#endif
  return EXIT_SUCCESS;
}
